/*
 * This file is part of nads
 * Copyright (c) 2003 ECSC Ltd.
 * Author: Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
 *
 * Here we have the program initialisation an main loop.
*/

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#include <nads.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

/* Order of IPC tokens */
#define PARAM_PROTO	0
#define PARAM_SRC	1
#define PARAM_DST	2
#define PARAM_PORT	3
#define PARAM_METHOD	4
#define PARAM_URL	5
#define PARAM_MAX	6

/* Input buffer */
static char buf[8192];

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"webserver",	1, 0, 'p'},
	{"expressions",	1, 0, 'e'},
	{"help",	0, 0, 'h'},
	{NULL, 0, 0, 0}
};
#endif

/* usage:
 * @cmd: command name we were invoked with
 * @code: exit code
 *
 * Prints the usage string then exits.
 */
static __attribute__((noreturn)) void usage(char *cmd, int code)
{
	fprintf(stderr, "Usage: %s [-w webserver]\n", cmd);
	fprintf(stderr, "Squid ACL helper to block web attacks\n\n");
	fprintf(stderr, "  --webserver, -w      Select default webserver emulation\n");
	fprintf(stderr, "  --expressions, -e    Load old-style regular expressions\n");
	fprintf(stderr, "  --help, -h           This message\n\n");
	/* TODO: List webservers available */
	exit(code);
}

/* check_args:
 * @argc: argument count
 * @argv: arguments
 *
 * Parse command line options. May never return.
 */
static void check_args(int argc, char **argv)
{
	int c;
	const char *ws=NULL;
	char *cmd=argv[0];

#ifdef HAVE_GETOPT_LONG
	while( (c=getopt_long(argc, argv, "w:e:h?", opts, NULL)) != -1 )
#else
	while( (c=getopt(argc, argv, "w:e:h?")) != -1 )
#endif
	{
		switch ( c ) {
		case 'w':
			if ( ws )
				usage(cmd, 1);
			ws = optarg;
			break;
		case 'e':
			break;
		case 'h':
		case '?':
			usage(cmd, 0);
			break;
		}
	}

	/* Set webserver */
	if ( ws && !nads_default_webserver(ws) ) {
		fprintf(stderr, "%s: %s: %s\n", cmd, ws, nads_errstr());
		exit(1);
	}
}

/* parse_req:
 * @buf: IPC buffer
 *
 * Parses the squid IPC in "%PROTO %SRC %DST %PORT %METHOD %PATH"
 * format. Thr request is then passed on to be checked and normalized.
 */
static void parse_req(char *buf)
{
	char *tok[PARAM_MAX];
	char *cur;
	int num=0;
	int state=0;
	struct nads n = {.host = 0, .port = 0};
	unsigned long int ret;
	char *end=NULL;

	for(cur=buf; *cur; cur++) {
		if ( state== 0 ) {
			if ( *cur!=' ' ) {
				tok[num++]=cur;
				state=1;
				if ( num >= PARAM_MAX )
					break;
			}
		}else{
			if ( *cur==' ' ) {
				*cur=0;
				state=0;
			}
		}
	}

	if ( num != PARAM_MAX )
		goto ipc_error;

	/* Allow all non-http (we can't analyze that) */
	if ( strcmp(tok[PARAM_PROTO], "http") ) {
		printf("OK\n");
		return;
	}

	n.buf = tok[PARAM_URL];
	n.host = tok[PARAM_DST];

	/* Convert port number */
	ret = strtoul(tok[PARAM_PORT], &end, 10);

	if ( end == tok[PARAM_PORT] )
		goto ipc_error;

	if ( ret == ULONG_MAX && errno == ERANGE )
		goto ipc_error;

	if ( ret & ~0xffffUL )
		goto ipc_error;

	n.port = (u_int16_t)ret;

	if ( !nads_normalize(&n) ) {
		printf("ERR error=\"%s\"\n", nads_errstr());
		return;
	}

	/* TODO: match rules etc. */

	printf("OK\n");
	return;
ipc_error:
	printf("ERR error=\"IPC error\"\n");
	return;
}

/* main:
 * @argc:
 * @argv:
 *
 * Program entry point
 */
int main(int argc, char **argv) 
{
	char *end;
	int state=0;

	check_args(argc, argv);

	while ( fgets(buf, sizeof(buf), stdin) ) {
		end = strchr(buf, '\n');

		/* Continuation of a long line */
		if ( state==1 ) {
			if ( end ) {
				printf("ERR error=\"IPC request too long\"\n");
				state=0;
			}
			continue;
		}

		/* Start of a long line */
		if ( !end ) {
			state = 1;
			continue;
		}

		/* OK to handle URL */
		*end=0;
		parse_req(buf);
		fflush(stdout);
	}

	return 0;
}
