/*
 * This file is part of nads
 * Copyright (c) 2003 ECSC Ltd.
 * Author: Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
 *
 * Because different webservers require different normalization due to
 * non RFC-strict implementations and abiguities in the standards. This
 * file keeps track of which one the user has selected.
 *
 * TODO:
 *  o Need to add more emulators for other webservers
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <nads.h>
#include <string_hash.h>

#define IPHASH_SIZE 128

struct emulator {
	const char *name;
	unsigned long flags;
};

struct webserver {
	iphash_t hash;
	const char *host;
	u_int16_t port;
	struct emulator *emu;
	struct webserver *next;
};

/* Emulation database */
static struct emulator emulators[]={
	{"IIS",		NADS_WS_MS_UTF16 | NADS_WS_PERCENT_HACK |
			NADS_WS_DOUBLE_HEX },
	{"Apache",	NADS_WS_CASE_SENSITIVE},
};

/* Default emulator */
static struct webserver default_ws = {
	.hash = 0,
	.host = NULL,
	.port = 0,
	.emu = &emulators[0],
};

/* IP:Port -> Webserver mapping */
static struct webserver *webservers[IPHASH_SIZE];

/* Currently selected webserver */
static struct webserver *current = &default_ws;

/* find_emulator:
 * @name: Name of emulator to find
 *
 * Find an emulator based on it's name
 *
 * Return value: Pointer to the emulator, or NULL on error
 */
static struct emulator *find_emulator(const char *name)
{
	int num=sizeof(emulators)/sizeof(*emulators);
	int i;

	for(i=0; i<num; i++) {
		if ( !strcasecmp(emulators[i].name, name) ) {
			return &emulators[i];
		}
	}

	return NULL;
}

/* nads_default_webserver:
 * @name: Name of the webserver to protect
 *
 * Selects the webserver to use for any addresses we don't
 * know about.
 *
 * Return value: non-zero for success, zero for error.
 */
int nads_default_webserver(const char *name)
{
	struct emulator *e;

	e = find_emulator(name);
	if ( e == NULL ) {
		nads_errcode = NADS_ERR_NO_EMULATION;
		return NADS_FAIL;
	}

	default_ws.emu = e;
	nads_errcode = NADS_ERR_SUCCESS;
	return NADS_OK;
}

/* nads_ws:
 * @x: the feature(s) we are checking for
 *
 * Returns true if ALL the requested features bits are set
 * by the currently selected webserver.
 *
 * Return value: 1 for true, 0 for false
 */
int __hidden nads_ws(unsigned long x)
{
	return ( (current->emu->flags & x) != 0 );
}

/* nads_webserver:
 * @host: Host name of web server
 * @port: TCP port it is listening on
 * @name: Name of the webserver to emulate
 *
 * Maps a webserver emulation type on to an ip:port tuple
 *
 * Return value: non-zero for success, zero for error.
 */
int nads_webserver(const char *host, u_int16_t port, const char *name)
{
	struct webserver *w;
	struct emulator *e;

	e = find_emulator(name);
	if ( e == NULL ) {
		nads_errcode = NADS_ERR_NO_EMULATION;
		return 0;
	}

	w = calloc(1, sizeof(*w));
	if ( w == NULL ) {
		nads_errcode = NADS_ERR_OOM;
		return 0;
	}

	w->emu = e;
	w->hash = webserver_hash(host, port);
	w->host = host;
	w->port = port;
	w->next = webservers[w->hash % IPHASH_SIZE];
	webservers[w->hash % IPHASH_SIZE] = w;

	nads_errcode = NADS_ERR_SUCCESS;
	return 1;
}

/* select_webserver:
 * @host: Host name of server to select
 * @port: TCP port it is listening on
 *
 * Selects a webserver for comparisons, used from nads_normalize
 */
void __hidden select_webserver(const char *host, u_int16_t port)
{
	struct webserver *w;
	iphash_t h;

	if ( (current->port == port)  &&
		!strcasecmp(current->host, host) )
		return;

	if ( host == NULL && port == 0 ) {
		current = &default_ws;
		return;
	}

	h = webserver_hash(host, port);

	for(w=webservers[h % IPHASH_SIZE]; w; w=w->next) {
		if ( (h == w->hash) &&
			(port == w->port) &&
			!strcasecmp(host, w->host) ) {
			current = w;
			return;
		}
	}

	current = &default_ws;
}
